const math = @import("./math.zig");
const meta = @import("./meta.zig");
const stack = @import("./stack.zig");
const testing = @import("./testing.zig");

///
/// Dynamic memory allocation interface.
///
pub const Allocator = union (enum) {
    bound: struct {
        alloc: fn (usize) ?[*]u8,
        dealloc: fn ([*]u8) void,
    },

    arena: struct {
        buffer: []u8,
        cursor: usize = 0,
    },

    ///
    /// [MakeError.OutOfMemory] is used to indicate there is not enough memory available for a given
    /// operation.
    ///
    pub const MakeError = error {
        OutOfMemory,
    };

    ///
    /// Frees `allocation` using `allocator`.
    ///
    pub fn free(allocator: *Allocator, allocation: anytype) void {
        switch (@typeInfo(@TypeOf(allocation))) {
            .Pointer => |pointer| if (pointer.size == .Slice)
                @compileError("`allocation` cannot be a slice"),

            else => @compileError("`allocation` must be a pointer"),
        }

        if (@typeInfo(@TypeOf(allocation)) != .Pointer)
            @compileError("`allocation` must be a pointer");

        // TODO: Implement arena de-allocation.
        switch (allocator.*) {
            .bound => |bound| bound.dealloc(@ptrCast([*]u8, allocation)),
            .arena => {},
        }
    }

    ///
    /// Attempts to allocate `size` number of `Element`s using `allocator`.
    ///
    /// Returns the allocation or a [MakeError] if it failed.
    ///
    pub fn make(allocator: *Allocator, comptime Element: type, size: usize) MakeError![*]Element {
        switch (allocator.*) {
            .bound => |bound| return @ptrCast([*]Element, @alignCast(@alignOf(Element),
                bound.alloc(@sizeOf(Element) * size) orelse return error.OutOfMemory)),

            .arena => |*stack| {
                defer stack.cursor += size;

                return @ptrCast([*]Element, @alignCast(@alignOf(Element), stack.buffer.ptr));
            },
        }
    }
};

///
/// Closure that captures a reference to readable resources like block devices, memory buffers,
/// network sockets, and more.
///
pub const Reader = meta.Function(@sizeOf(usize), []u8, usize);

///
/// Returns a state machine for lazily computing all `Element` components of a given source input
/// that match a delimiting pattern.
///
pub fn Spliterator(comptime Element: type) type {
    return struct {
        source: []const Element,
        delimiter: []const Element,

        const Self = @This();

        ///
        /// Returns `true` if there is more data to be processed, otherwise `false`.
        ///
        pub fn hasNext(self: Self) bool {
            return (self.source.len != 0);
        }

        ///
        /// Iterates on `self` and returns the next view of [Spliterator.source] that matches
        /// [Spliterator.delimiter], or `null` if there is no more data to be processed.
        ///
        pub fn next(self: *Self) ?[]const Element {
            if (!self.hasNext()) return null;

            if (self.delimiter.len == 0) {
                defer self.source = self.source[self.source.len .. self.source.len];

                return self.source[0 .. self.source.len];
            }

            if (findFirstOf(Element, self.source, self.delimiter, struct {
                fn testEquality(this: Element, that: Element) bool {
                    return this == that;
                }
            }.testEquality)) |head| {
                defer self.source = self.source[(head + self.delimiter.len) .. self.source.len];

                return self.source[0 .. head];
            }

            defer self.source = self.source[self.source.len .. self.source.len];

            return self.source;
        }
    };
}

test "Spliterator of string literals" {
    // Empty source.
    {
        var spliterator = Spliterator(u8){
            .source = "",
            .delimiter = " ",
        };

        try testing.expect(!spliterator.hasNext());
    }

    // Empty delimiter.
    {
        var spliterator = Spliterator(u8){
            .source = "aaa",
            .delimiter = "",
        };

        try testing.expect(spliterator.hasNext());
        try testing.expect(equals(u8, spliterator.next().?, "aaa"));
        try testing.expect(!spliterator.hasNext());
    }

    // Single-character delimiter.
    {
        var spliterator = Spliterator(u8){
            .source = "single.character.separated.hello.world",
            .delimiter = ".",
        };

        const components = [_][]const u8{"single",
            "character", "separated", "hello", "world"};

        var index = @as(usize, 0);
        const components_tail = components.len - 1;

        while (spliterator.next()) |split| : (index += 1) {
            try testing.expect(spliterator.hasNext() == (index < components_tail));
            try testing.expect(equals(u8, split, components[index]));
        }

        try testing.expect(!spliterator.hasNext());
    }

    // Multi-character delimiter.
    {
        var spliterator = Spliterator(u8){
            .source = "finding a needle in a needle stack",
            .delimiter = "needle",
        };

        const components = [_][]const u8{"finding a ", " in a ", " stack"};
        var index = @as(usize, 0);
        const components_tail = components.len - 1;

        while (spliterator.next()) |split| : (index += 1) {
            try testing.expect(spliterator.hasNext() == (index < components_tail));
            try testing.expect(equals(u8, split, components[index]));
        }

        try testing.expect(!spliterator.hasNext());
    }
}

///
/// Closure that captures a reference to writable resources like block devices, memory buffers,
/// network sockets, and more.
///
pub const Writer = meta.Function(@sizeOf(usize), []const u8, usize);

///
/// Returns a sliced reference of the raw bytes in `pointer`.
///
/// **Note** that passing a slice will convert it to a byte slice.
///
pub fn bytesOf(pointer: anytype) switch (@typeInfo(@TypeOf(pointer))) {
    .Pointer => |info| if (info.is_const) []const u8 else []u8,
    else => @compileError("`pointer` must be a pointer type"),
} {
    const Pointer = @TypeOf(pointer);
    const pointer_info = @typeInfo(Pointer).Pointer;

    switch (pointer_info.size) {
        .Many => @compileError("`pointer` cannot be an unbound pointer type"),
        .C => @compileError("`pointer` cannot be a C-style pointer"),

        .One => return @ptrCast(if (pointer_info.is_const) [*]const u8
                else [*]u8, pointer)[0 .. @sizeOf(Pointer)],

        .Slice => return @ptrCast(if (pointer_info.is_const) [*]const u8 else
            [*]u8, pointer.ptr)[0 .. (@sizeOf(Pointer) * pointer.len)],
    }
}

test "Bytes of types" {
    var foo: u32 = 10;

    try testing.expect(bytesOf(&foo)[0] == 0x0a);
}

///
/// Compares `this` to `that`, returning the difference between the first byte deviation in the two
/// sequences, otherwise `0` if they are identical.
///
pub fn compareBytes(this: []const u8, that: []const u8) isize {
    const range = math.min(usize, this.len, that.len);
    var index: usize = 0;

    while (index < range) : (index += 1) {
        const difference = (this[index] - that[index]);

        if (difference != 0) return difference;
    }

    return (@intCast(isize, this.len) - @intCast(isize, that.len));
}

test "Compare bytes" {
    try testing.expect(compareBytes(&.{69, 42, 0}, &.{69, 42, 0}) == 0);
    try testing.expect(compareBytes(&.{69, 42, 11}, &.{69, 42}) == 1);
    try testing.expect(compareBytes(&.{69, 42}, &.{69, 42, 11}) == -1);
}

///
/// Copies the contents of `source` into `target`
///
pub fn copy(comptime Element: type, target: []Element, source: []const Element) void {
    for (source) |element, index| target[index] = element;
}

test "Copy data" {
    var buffer = [_]u32{0} ** 20;
    const data = [_]u32{3, 20, 8000};

    copy(u32, &buffer, &data);

    for (data) |datum, index| try testing.expect(buffer[index] == datum);
}

///
/// Returns `true` if `this` is the same length and contains the same data as `that`, otherwise
/// `false`.
///
pub fn equals(comptime Element: type, this: []const Element, that: []const Element) bool {
    if (this.len != that.len) return false;

    var index: usize = 0;

    while (index < this.len) : (index += 1) if (this[index] != that[index]) return false;

    return true;
}

test "Check memory is equal" {
    const bytes_sequence = &.{69, 42, 0};

    try testing.expect(equals(u8, bytes_sequence, bytes_sequence));
    try testing.expect(!equals(u8, bytes_sequence, &.{69, 42}));
}

///
/// Searches for the first instance of an `Element` equal to `needle` in `haystack`, returning its
/// index or `null` if nothing was found.
///
pub fn findFirst(comptime Element: type, haystack: []const Element,
    needle: Element, comptime testEquality: fn (Element, Element) bool) ?usize {

    for (haystack) |element, index| if (testEquality(element, needle)) return index;

    return null;
}

test "Find first of element" {
    const haystack = &.{"", "", "foo"};

    const testEquality = struct {
        fn testEquality(this: []const u8, that: []const u8) bool {
            return equals(u8, this, that);
        }
    }.testEquality;

    try testing.expect(findFirst([]const u8, haystack, "foo", testEquality).? == 2);
    try testing.expect(findFirst([]const u8, haystack, "bar", testEquality) == null);
}

///
/// Searches for the first instance of an `Element` sequence equal to the contents of `needle` in
/// `haystack`, returning the starting index or `null` if nothing was found.
///
pub fn findFirstOf(comptime Element: type, haystack: []const Element,
    needle: []const Element, comptime testEquality: fn (Element, Element) bool) ?usize {

    var head: usize = 0;
    const tail = (haystack.len - needle.len);

    walk_haystack: while (head <= tail) : (head += 1) {
        for (needle) |element, index|
            if (!testEquality(haystack[head + index], element)) continue: walk_haystack;

        return head;
    }

    return null;
}

test "Find first of sequence" {
    const haystack = &.{"foo", "bar", "baz"};

    const testEquality = struct {
        fn testEquality(this: []const u8, that: []const u8) bool {
            return equals(u8, this, that);
        }
    }.testEquality;

    try testing.expect(findFirstOf([]const u8, haystack, &.{"bar", "baz"}, testEquality).? == 1);
    try testing.expect(findFirstOf([]const u8, haystack, &.{"baz", "bar"}, testEquality) == null);
}

///
/// Returns a deterministic hash code compiled from each byte in `bytes`.
///
/// **Note** that this operation has `O(n)` time complexity.
///
pub fn hashBytes(bytes: []const u8) usize {
    var hash = @as(usize, 5381);

    for (bytes) |byte| hash = ((hash << 5) + hash) + byte;

    return hash;
}

test "Hashing bytes" {
    const bytes_sequence = &.{69, 42, 0};

    try testing.expect(hashBytes(bytes_sequence) == hashBytes(bytes_sequence));
    try testing.expect(hashBytes(bytes_sequence) != hashBytes(&.{69, 42}));
}

///
/// Swaps the `Data` in `this` with `that`.
///
pub fn swap(comptime Data: type, this: *Data, that: *Data) void {
    const temp = this.*;
    this.* = that.*;
    that.* = temp;
}

test "Data swapping" {
    var a: u64 = 0;
    var b: u64 = 1;

    swap(u64, &a, &b);

    try testing.expect(a == 1);
    try testing.expect(b == 0);
}

///
/// Returns a [Writer] that silently consumes all given data without failure and throws it away.
///
/// This is commonly used for testing or redirected otherwise unwanted output data that has to be
/// sent somewhere for whatever reason.
///
pub fn nullWriter() Writer {
    return Writer.capture(@as(usize, 0), struct {
        fn write(_: usize, buffer: []const u8) usize {
            return buffer.len;
        }
    }.write);
}

test "Null writing" {
    const sequence = "foo";

    try testing.expect(nullWriter().call(sequence) == sequence.len);
}
